/*
    Sundtek USB Khelper module
    Copyright (C) 2010 Sundtek Ltd.

	This is a simple module which is designed to work with one
	device to improve the data throughput on a dreambox STB
*/

#include <linux/version.h>
#include <linux/init.h>
#include <linux/list.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/wait.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/i2c.h>
#include <linux/usb.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/fcntl.h>
#include <linux/vmalloc.h>
#include "usbkhelper.h"
#include "usbkhelper-user.h"

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 35)
#define usb_buffer_alloc(dev, size, mem_flags, dma) usb_alloc_coherent(dev, size, mem_flags, dma)
#define usb_buffer_free(dev, size, addr, dma) usb_free_coherent(dev, size, addr, dma)
#endif
extern int bcm_xpt_clip_enqueue(void *ptr, u8 data, ssize_t size, int flag);

static struct usb_device_id usb_khelper_id_table[]={
	{ USB_DEVICE(0xeb1a, 0x51b2) },
	{ USB_DEVICE(0xeb1a, 0x51b3) },
	{ USB_DEVICE(0xeb1a, 0x51b4) },
	{ USB_DEVICE(0xeb1a, 0x51b5) },
	{ USB_DEVICE(0xeb1a, 0x51b6) },
	{ USB_DEVICE(0xeb1a, 0x51b7) },
	{ USB_DEVICE(0xeb1a, 0x51b8) },
	{ USB_DEVICE(0xeb1a, 0x51b9) },
	{ USB_DEVICE(0xeb1a, 0x51ba) },
	{ USB_DEVICE(0xeb1a, 0x51bb) },
	{ USB_DEVICE(0xeb1a, 0x51bc) },
	{ USB_DEVICE(0xeb1a, 0x51bd) },
	{ USB_DEVICE(0xeb1a, 0x51be) },
	{ USB_DEVICE(0xeb1a, 0x8175) },
	{ USB_DEVICE(0x2659, 0x1208) },
	{}
};

MODULE_DEVICE_TABLE(usb, usb_khelper_id_table);

static unsigned long usbkhelper_devices; /* map */
static struct list_head usbkhelper_devlist;

static unsigned int delay=0;
module_param(delay, int, 0444);
MODULE_PARM_DESC(delay, "delay before returning from read in milliseconds (default 0ms)");
    
static int __major;

#define BUFFER_SIZE 1880000

static struct rb_info *rb_init(void);
static struct rb_info *rb_init(void) {
        struct rb_info *rb = kmalloc(sizeof(struct rb_info), GFP_KERNEL);
        rb->left = rb->right = rb->pos = 0;
        rb->bufsize = BUFFER_SIZE;
        rb->buffer = vmalloc_32(BUFFER_SIZE*sizeof(u8));
        if (rb->buffer == NULL)
            return 0;
        spin_lock_init(&rb->__lock);
        return rb;
};

void rb_free(struct rb_info **rb) {
	vfree((*rb)->buffer);
        kfree((*rb));
}

static inline u32 rb_get_buffer_size(struct rb_info *rb) {
        if (rb->left > rb->right)
                return (rb->bufsize - rb->left + rb->right);
        else if(rb->left < rb->right)
                return (rb->right - rb->left);
        else if(rb->left == rb->right)
                return 0;
        return 0;
}

static void rb_reset_buffer(struct rb_info *rb) {
        unsigned long flags;
        spin_lock_irqsave(&rb->__lock, flags);
        rb->left = rb->right = rb->pos = 0;
        spin_unlock_irqrestore(&rb->__lock, flags);
}

static inline u32 rb_free_buffer_size(struct rb_info *rb) {
        u32 cur_buffsize = rb_get_buffer_size(rb);
        if (rb->bufsize - cur_buffsize > 0)
                return rb->bufsize - cur_buffsize - 1;
        return rb->bufsize - 1;
}
static int rb_write_data(struct usbkhelper_info *info, struct rb_info *rb, u8 *data, ssize_t size) {
        int len = size;
        int free;
	int used = 0;
        unsigned long flags;
        spin_lock_irqsave(&rb->__lock, flags);
	/* thanks to dream multimedia for implementing this optimization to the dreambox based systems */
	if (info->use_acceleration && info->accel_ptr && info->bcm_xpt_clip_enqueue) {
		int ret;
		ret = info->bcm_xpt_clip_enqueue(info->accel_ptr, data, size, -2);
		if (ret != 0) {
			if (printk_ratelimit()) {
				printk("usbkhelper: accel enqueue failed (buffer full)!\n");
			}
		} else if (ret != size) {
			if (printk_ratelimit()) {
				printk("usbkhelper: accel enqueue failed (not enough space in buffer)!\n");
			}
		}
	} else {
		free = rb_free_buffer_size(rb);

		if (size > free)
			size = free - free % sizeof(u32); 

		if (free == 0) {
			if(printk_ratelimit())
				printk("dropping data!\n");
			spin_unlock_irqrestore(&rb->__lock, flags);
			return 0;
		}
		if (rb->right + size >= rb->bufsize - 1) {
			len = rb->bufsize - rb->right;
			memcpy(&rb->buffer[rb->right], data, len);
			memcpy(rb->buffer, data+len, size - len);
			rb->right = size - len;
		} else {
			memcpy(&rb->buffer[rb->right], data, size);
			rb->right += size;
		}
		/* unlock */

		used = rb_get_buffer_size(rb);
	}
        spin_unlock_irqrestore(&rb->__lock, flags);
        return used;
}

static int rb_read_data(struct rb_info *rb, u8 *buffer, ssize_t size) {
        int len;
        int bufsize;
        unsigned int left;
        unsigned long flags;
        /* lock */
        spin_lock_irqsave(&rb->__lock, flags);

        if (rb_get_buffer_size(rb) == 0) {
                /* unlock */
                spin_unlock_irqrestore(&rb->__lock, flags);
                return 0;
        }

        bufsize = rb_get_buffer_size(rb);
        if (size > bufsize)
                size = bufsize;

        size -= size % 188; /* alignment */

        if (rb->left + size >= rb->bufsize -1) {
                len = rb->bufsize - rb->left;
                left = rb->left;
                rb->left = size - len;
                spin_unlock_irqrestore(&rb->__lock, flags);
                copy_to_user(buffer, &rb->buffer[left], len);
                copy_to_user(&buffer[len], rb->buffer, size - len);
        } else {
                left = rb->left;
                rb->left += size;
                spin_unlock_irqrestore(&rb->__lock, flags);
                copy_to_user(buffer, &rb->buffer[left], size);
        }
        /* unlock */
        return size;
}


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
static void usbkhelper_isocirq(struct urb *urb, struct pt_regs *regs)
#else
static void usbkhelper_isocirq(struct urb *urb)
#endif
{
    struct usbkhelper_info *info = urb->context;
    u8 *cp;
    int cnt=0;
    int i;
    if (info->bulk_transfer) {
	    cnt+=rb_write_data(info, info->__usbkhelper_rb, urb->transfer_buffer, urb->actual_length);
    } else {
	    for(i=0;i<urb->number_of_packets;i++){
		    int length=urb->iso_frame_desc[i].actual_length;
		    cp=(unsigned char *) urb->transfer_buffer + urb->iso_frame_desc[i].offset;

		    if(!length)
			    continue;


		    if (urb->iso_frame_desc[i].status)
			    continue;

		    cnt+=rb_write_data(info, info->__usbkhelper_rb, cp, length);

	    }
    }
    urb->status = 0;
 

    if (info->stop_transfer==0)
	    if(usb_submit_urb(urb, GFP_ATOMIC))
		    printk("unable to submit urb\n");

    if (cnt>18800) {
	wake_up_interruptible(&info->wait);
    }
}

static int usbkhelper_isoc_deinit(struct usbkhelper_info *info)
{
    int i;
    info->stop_transfer=1;
    printk("usbkhelper: stopping data transfer\n");
    for(i=0;i<USBKHELPER_BUFS;i++){
        usb_kill_urb(info->urb[i]);
        usb_free_urb(info->urb[i]);
        info->urb[i]=NULL;
    }
    wake_up_interruptible(&info->wait);
    return 0;
}

static int usbkhelper_isoc_init(struct usbkhelper_info *info)
{
    int i;
    int errCode;
    int sb_size;

    printk(KERN_INFO "usbkhelper: starting data transfer\n");
    info->stop_transfer=0;
    if (info->bulk_transfer) {
	    sb_size= USBKHELPER_NUM_PACKETS * 64; //512;
    } else {
	    sb_size=USBKHELPER_NUM_PACKETS * USBKHELPER_MAX_PACKET_SIZE;
    }

    for(i=0;i<USBKHELPER_BUFS;i++){
        struct urb *urb;
        int j,k;
	if (info->transfer_buffer[i] == NULL)
		info->transfer_buffer[i]=kmalloc(sb_size,GFP_ATOMIC);

        if(!info->transfer_buffer[i]) {
	    printk(KERN_INFO "unable to allocate transfer memory\n");
            return -ENOMEM;
	}

        memset(info->transfer_buffer[i],0x80,sb_size);
        urb = usb_alloc_urb(USBKHELPER_NUM_PACKETS,GFP_ATOMIC);

        if(urb){
                if (info->bulk_transfer) {
			urb->dev=info->__udev;
			urb->context=info;
			urb->pipe=usb_rcvbulkpipe(info->__udev, 0x84);
                        urb->transfer_flags = 0;
			urb->transfer_buffer = info->transfer_buffer[i];
                        urb->interval = 0;
			urb->complete = usbkhelper_isocirq;
			urb->number_of_packets = USBKHELPER_NUM_PACKETS;
			urb->transfer_buffer_length = sb_size;
                        urb->transfer_buffer = info->transfer_buffer[i];
			info->urb[i]=urb;

                } else {
			urb->dev=info->__udev;
			urb->context=info;
			urb->pipe=usb_rcvisocpipe(info->__udev, 0x84);
			urb->transfer_flags = URB_ISO_ASAP;
			urb->transfer_buffer = info->transfer_buffer[i];
			urb->interval=1;
			urb->complete = usbkhelper_isocirq;
			urb->number_of_packets = USBKHELPER_NUM_PACKETS;
			urb->transfer_buffer_length = sb_size;
			for(j=k=0; j<USBKHELPER_NUM_PACKETS;j++,k+=USBKHELPER_MAX_PACKET_SIZE){
				urb->iso_frame_desc[j].offset = k;
				urb->iso_frame_desc[j].length=USBKHELPER_MAX_PACKET_SIZE;
			}
			info->urb[i]=urb;
                }

        } else {
	    printk(KERN_INFO "unable to allocate transfer memory\n");
            return -ENOMEM;
	}

    }
    for(i=0;i<USBKHELPER_BUFS;i++){
        errCode = usb_submit_urb(info->urb[i], GFP_ATOMIC);
        if (errCode){
	    printk(KERN_INFO"submit problem %d\n", errCode);
            usbkhelper_isoc_deinit(info);
            return errCode;
        }
    }
    return 0;
}


static int usb_khelper_open(struct inode *inode, struct file *file) {
        struct usbkhelper_info *info = NULL;
        struct list_head *list;
	int minor = iminor(inode);

        /* TODO: support multiple devices here */
        list_for_each(list, &usbkhelper_devlist) {
                info = list_entry(list, struct usbkhelper_info, __usbkhelper_devlist);
		if (info->__id == minor)
			break;
        }
        
        if (info == NULL) {
		printk(KERN_INFO "no device information found");
                return -1;
	}
	
	if (info->shutdown==1)
		return -1;

	if (info->users != 0) {
		printk(KERN_INFO "device is already opened %d\n", info->users);
                return -EBUSY;
	}
        info->users++;

        file->private_data = info;

        rb_reset_buffer(info->__usbkhelper_rb);
        return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
static int usb_khelper_ioctl(struct inode *inode, struct file *filp, unsigned int cmd, unsigned long arg) {
#else
static long usb_khelper_ioctl(struct file *filp, unsigned int cmd, unsigned long arg) {
#endif
	void *parg = (void*)arg;
	int ret = 0;
        struct usbkhelper_info *info = NULL;
        info = filp->private_data;
	switch (cmd) {
	case KHELPER_GET_DEVICE:
	{
		struct khelper_device kdev;
		kdev.bus    = info->busid;
		kdev.device = info->devid;
		ret = copy_to_user(parg, &kdev, sizeof(struct khelper_device));
		if (ret) {
			printk(KERN_INFO "error copying\n");
			return -1;
		}
		break;
	}
	case KHELPER_USE_ACCELERATION:
	{
		u8 buf = 0;
		ret = copy_from_user(&buf, parg, sizeof(u8));
		if (ret) {
			printk(KERN_INFO "error copying from user\n");
			return -1;
		}

		switch(buf) {
		case 0: /* disable */
			info->use_acceleration=0;
			break;
		case 1: /* enable */
			info->use_acceleration=1;
			break;
		default:
			printk(KERN_INFO "argument error\n");
			return -1;
		}
		break;
	}
	case KHELPER_START_TRANSFER:
	{
		mutex_lock(&info->lockcmd);
		info->transfer_started = 1;
		usb_set_interface(info->__udev, 0, 1);
		ret = usbkhelper_isoc_init(info);
		if (ret != 0) {
			printk(KERN_INFO "enabling transfer failed");
			info->transfer_started = 0;
			mutex_unlock(&info->lockcmd);
			return -1;
		}
		mutex_unlock(&info->lockcmd);
		break;
	}
	case KHELPER_STOP_TRANSFER:
	{
		mutex_lock(&info->lockcmd);
		if (info->transfer_started == 1) {
			info->transfer_started = 0;
			usbkhelper_isoc_deinit(info);
			usb_set_interface(info->__udev, 0, 0);
		}
		mutex_unlock(&info->lockcmd);
		break;
	}
	case KHELPER_SET_PTR:
	{
		info->accel_ptr = (void*)arg;
		break;
	}
	default:
		printk(KERN_INFO "unknown ioctl\n");
	}
	return ret;
}

static ssize_t usb_khelper_read(struct file *fp, char __user *buf, size_t count, loff_t *f_pos) {
        struct usbkhelper_info *info = NULL;
        ssize_t retval=0;
	int ret;
        info = fp->private_data;
	mutex_lock(&info->lockcmd);
	if (info->transfer_started == 0) {
		info->transfer_started = 1;
		usb_set_interface(info->__udev, 0, 1);
		ret = usbkhelper_isoc_init(info);
		if (ret != 0) {
			info->transfer_started = 0;
			info->users--;
			printk(KERN_INFO "enabling transfer failed");
			mutex_unlock(&info->lockcmd);
			return -1;
		}
	}
	mutex_unlock(&info->lockcmd);
	if (delay>200) 
		delay=200;
	if (delay>0)
		msleep(delay);
//	retval = rb_read_data(info->__usbkhelper_rb, buf, count);
	if (fp->f_flags & O_NONBLOCK) {
		retval=rb_read_data(info->__usbkhelper_rb, buf, count);
		ret=0;
	} else {
		ret = wait_event_interruptible(info->wait, (retval=rb_read_data(info->__usbkhelper_rb, buf, count))>0);
	}
	if (ret<0) {
		return 0;
	}
	return retval;
}

static int usb_khelper_release(struct inode *inode, struct file *file) {
	struct usbkhelper_info *info;
	info = file->private_data;
	wake_up_interruptible(&info->wait);
	mutex_lock(&info->lockcmd);
	if (info->transfer_started == 1) {
		info->transfer_started = 0;
		usbkhelper_isoc_deinit(info);
		usb_set_interface(info->__udev, 0, 0);
	}
	mutex_unlock(&info->lockcmd);
	info->users--;
	if (info->users == 0)
		wake_up(&info->disconnect);
	return 0;
}

static unsigned int usb_khelper_poll(struct file *filp, poll_table *wait) {
	struct usbkhelper_info *info;
	unsigned int mask = 0;
	info = filp->private_data;
	poll_wait(filp, &info->wait, wait);
	if (rb_get_buffer_size(info->__usbkhelper_rb)>188) {
		mask |= POLLIN|POLLPRI|POLLRDNORM;
	}
	return mask;
}

static struct file_operations usbkhelper_fops = {
	.owner = THIS_MODULE,
	.open  = usb_khelper_open,
	.read  = usb_khelper_read,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,38)
	.ioctl = usb_khelper_ioctl,
#else
	.unlocked_ioctl = usb_khelper_ioctl,
#endif
	.poll  = usb_khelper_poll,
	.release = usb_khelper_release,
};

static int usb_khelper_probe(struct usb_interface *interface, const struct usb_device_id *uid) {
	int max_alt;
	int i;
	int id;
	struct usbkhelper_info *info=NULL;
	max_alt = interface->num_altsetting;
	if (max_alt != 2 && max_alt != 8)
		return -1;

	info = kmalloc(sizeof(struct usbkhelper_info), GFP_KERNEL);
	memset(info, 0x0, sizeof(struct usbkhelper_info));

	id = find_first_zero_bit(&usbkhelper_devices, sizeof(u32));
	usbkhelper_devices |= (1<<id);
	info->__id = id;
	info->__udev = usb_get_dev(interface_to_usbdev(interface));

	info->bcm_xpt_clip_enqueue = (void*)symbol_request(bcm_xpt_clip_enqueue);
	if (info->bcm_xpt_clip_enqueue) {
		printk(KERN_INFO"usbkhelper: dreambox acceleration detected\n");
	} else {
		printk(KERN_INFO"usbkhelper: dreambox acceleration not detected\n");
	}

	mutex_init(&info->lockcmd);
	info->busid = info->__udev->bus->busnum;
	info->devid = info->__udev->devnum;
	//printk(KERN_INFO"usbkhelper: %03d/%03d\n", info->busid, info->devid);

	switch (info->__udev->speed) {
		case USB_SPEED_HIGH:
			printk(KERN_INFO"usbkhelper: device is attached to a USB 2.0 bus\n");
			break;
		case USB_SPEED_FULL:
			printk(KERN_INFO"usbkhelper: setting up device on a USB 1.1 bus\n");
			/* passthrough */
		default:
			printk(KERN_INFO"usbkhelper: your device won't work properly when\n");
			printk(KERN_INFO"usbkhelper: not attached to a USB 2.0 highspeed bus\n");
	}


	for(i=0;i<interface->altsetting[0].desc.bNumEndpoints;i++) {
		if (interface->altsetting[0].endpoint[i].desc.bEndpointAddress == 0x84 &&
			(interface->altsetting[0].endpoint[i].desc.bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) == USB_ENDPOINT_XFER_BULK) {
			info->bulk_transfer = 1;	
			break;
		}
	}
	if (info->bulk_transfer == 1) {
		printk(KERN_INFO "usbkhelper: using bulk transfer\n");
	} else {
		printk(KERN_INFO "usbkhelper: using isoc transfer\n");
	}

	snprintf(info->__devname, 50, "ngdvr%d", id);
	printk(KERN_INFO "usbkhelper: registered /dev/%s\n", info->__devname);

	init_waitqueue_head(&info->wait);
	init_waitqueue_head(&info->disconnect);

	info->__class = class_create(THIS_MODULE, info->__devname);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
	info->__cdev  = class_device_create(info->__class, NULL, MKDEV(__major, info->__id), NULL, info->__devname);
#elif LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 27)
	info->__cdev = device_create_drvdata(info->__class, NULL,
			MKDEV(__major, info->__id), NULL, info->__devname);
#else
	info->__cdev = device_create(info->__class, NULL,
			MKDEV(__major, info->__id), NULL, info->__devname);
#endif

	printk(KERN_INFO "usbkhelper: mknod /dev/ngdvr%d c %d %d\n", info->__id,  __major, info->__id);

	list_add_tail(&info->__usbkhelper_devlist, &usbkhelper_devlist);


	info->__usbkhelper_rb = rb_init();
	info->__curr_rb = 0;

	usb_set_intfdata(interface, info);
	
	return 0;
}

static void usb_khelper_disconnect(struct usb_interface *interface) {
        struct usbkhelper_info *info;
	info = usb_get_intfdata(interface);
	printk(KERN_INFO "Disconnecting\n");
	info->shutdown=1;

	if (info->users>0) {
		printk(KERN_INFO "waiting for last user to disconnect\n");
		wait_event(info->disconnect, info->users==0);
	}

        usbkhelper_devices &= ~(1<<info->__id);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 26)
        class_device_destroy(info->__class, MKDEV(__major, info->__id));
#else
        device_destroy(info->__class, MKDEV(__major, info->__id));
#endif
        class_destroy(info->__class);
        rb_free(&info->__usbkhelper_rb);
        list_del(&info->__usbkhelper_devlist);

	if (info->bcm_xpt_clip_enqueue) {
		printk(KERN_INFO"usbkhelper: dreambox releasing bcm xpt clip enqueue\n");
		symbol_put_addr(info->bcm_xpt_clip_enqueue);
	}
        kfree(info);
}


static struct usb_driver usb_khelper_drv = {
#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 15)
	.owner = THIS_MODULE,
#endif
	.name = "usb_khelper",
	.probe = usb_khelper_probe,
	.disconnect = usb_khelper_disconnect,
	.id_table = usb_khelper_id_table
};

static int __init usb_khelper_init(void) {
	//struct usb_khelper *kh;
	int result;
	
	INIT_LIST_HEAD(&usbkhelper_devlist);
	__major = register_chrdev(0, "sundtek", &usbkhelper_fops);
	result = usb_register(&usb_khelper_drv); //KBUILD_MODNAME);
	if (result) {
		printk(KERN_INFO"usb khelper registration failed\n");
		return -1;
	}
	printk("Initializing Sundtek USB khelper\n");
	printk("Copyright (C) 2010 Sundtek Ltd\n");
	return 0;
}

static void __exit usb_khelper_exit(void) {
	printk("Releasing Sundtek USB khelper module\n");
	usb_deregister(&usb_khelper_drv);
        unregister_chrdev(__major, "usbkhelper");
}

module_init(usb_khelper_init);
module_exit(usb_khelper_exit);

MODULE_AUTHOR("Sundtek Ltd.");
MODULE_DESCRIPTION("USB Kernel Helper Module");
MODULE_LICENSE("GPL");
